"""Module for handling imports and exports with .csv files."""
from __future__ import annotations
from typing import TYPE_CHECKING
import csv

from . import validate_extension
if TYPE_CHECKING:
    from .. import Signal1


def validate_filename(filename: str) -> None:
    """Validates the name of the file.

    Parameters
    ----------
    filename : str
        Name of the file to check.
    """
    validate_extension(filename, "csv")


def export_signal1(filename: str, signal1: Signal1) -> None:
    """Exports the given one dimensional signal to the .csv file."""
    validate_filename(filename)
    with open(filename, "w+") as file:
        writer = csv.writer(file)
        writer.writerows(*signal1.unpack())


def import_signal1(filename: str) -> Signal1:
    """Imports a one dimensional signal from a .csv file."""
    validate_filename(filename)
    pass



import numpy as np
from scipy import signal

from ..sgn import Signal1
from ..config import H1_METHOD
from . import fourier, ifourier


def _char_function(t) -> float:
    return 1 / (np.pi * t)


def h1(signal1: Signal1, method=H1_METHOD) -> Signal1:
    return H1_METHODS[method](signal1)


def calculate_conv(signal1: Signal1) -> Signal1:
    output = signal1.clone()
    signal2 = Signal1.from_function(output.axis, _char_function)
    return output.convolute(signal2)


def calculate_fft(signal1: Signal1) -> Signal1:
    output = signal1.clone()
    signal2 = Signal1.from_function(output.axis, _char_function)
    self_fourier = fourier.f1(output)
    t_fourier = fourier.f1(signal2)
    return ifourier.if1(self_fourier * t_fourier)


def calculate_prod(signal1: Signal1) -> Signal1:
    output = signal1.clone()
    self_fourier = fourier.f1(output)
    # self_fourier = Fourier1(output).freq_shift()
    axis_len = len(output.axis)
    if axis_len % 2 == 0:
        h_values = 2 * np.ones(axis_len // 2 - 1)
    else:
        h_values = 2 * np.ones(axis_len // 2)
    h_values[0] = 1
    h_values[-1] = 1
    h_values = np.append(h_values, np.zeros(axis_len // 2 + 1))
    assert len(h_values) == axis_len
    # output.values = h_values
    # return InverseFourier1(output * self_fourier)
    output.values = output.values * self_fourier.values
    return ifourier.if1(output)


def calculate_scipy(signal1: Signal1) -> Signal1:
    output = signal1.clone()
    output.values = signal.hilbert(output.values)
    return output.imag_part()


H1_METHODS = {
    "conv": calculate_conv,
    "fft": calculate_fft,
    "prod": calculate_prod,
    "scipy": calculate_scipy,
}



"""
===========
Chirper GUI
===========

This is the app associated with Chirper, which allows the user to
visualize and manipulate signals from different sources. Currently, it
is able to fetch data from the following sources:
 - Microphone
"""


import numpy as np

from ..config import HERTZ, NOISE_TYPE
from . import Signal1


########################################################################################################################
# |||||||||||||||||||||||||||||||||||||||||||||||| Signal1 ||||||||||||||||||||||||||||||||||||||||||||||||||||||||||| #
########################################################################################################################

class SQUARE(Signal1):
    """Square signal"""

    def __init__(self, axis, freq, amp, rads=False, phase=0):
        """Generates a square signal centered at 0.

        Parameters
        ----------
        axis : array like
            Array for the axis.
        freq : float
            Frequency for the square wave.
        amp : float
            Amplitude of the wave.
        rads : bool
            Whether the frequency is given in radians or hertz, by default False.
        phase : float, optional
            Phase of the wave, by default 0.
        """
        super().__init__(axis, SQUARE._generate(axis, freq, amp, rads, phase))

    @staticmethod
    def _generate(axis, freq, amp, rads, phase):
        real_freq = freq if rads else 2 * np.pi * freq
        real_phase = phase if rads else 2 * np.pi * phase
        return amp * np.array(list(map(lambda x: int(x >= 0), np.sin(real_freq * axis + real_phase))))


class SIN(Signal1):
    """Sinusoidal signal"""

    def __init__(self, axis, freq, amp, hertz=HERTZ, phase=0):
        """Generates a sinusoidal signal centered at 0.

        Parameters
        ----------
        axis : array like
            Array for the axis.
        freq : float
            Frequency for the wave.
        amp : float
            Amplitude of the wave.
        hertz : bool, optional
            If True then the frequency is assumed to be in hertz, if not
            then it is in radians, by default config.HERTZ.
        phase : float, optional
            Phase of the wave, by default 0.
        """
        super().__init__(axis, SIN._generate(axis, freq, amp, hertz, phase))

    @staticmethod
    def _generate(axis, freq, amp, hertz, phase):
        real_freq = 2 * np.pi * freq if hertz else freq
        real_phase = 2 * np.pi * phase if hertz else phase
        return amp * np.sin(real_freq * axis + real_phase)


class COS(Signal1):
    """Cosine signal"""

    def __init__(self, axis, freq, amp, hertz=HERTZ, phase=0):
        """Generates a cosine signal centered at 0.

        Parameters
        ----------
        axis : array like
            Array for the axis.
        freq : float
            Frequency for the wave.
        amp : float
            Amplitude of the wave.
        hertz : bool, optional
            If True then the frequency is assumed to be in hertz, if not
            then it is in radians, by default config.HERTZ.
        phase : float, optional
            Phase of the wave, by default 0.
        """
        super().__init__(axis, COS._generate(axis, freq, amp, hertz, phase))

    @staticmethod
    def _generate(axis, freq, amp, hertz, phase):
        real_freq = 2 * np.pi * freq if hertz else freq
        real_phase = 2 * np.pi * phase if hertz else phase
        return amp * np.cos(real_freq * axis + real_phase)


class NOISE(Signal1):
    """Noise signal"""

    def __init__(self, axis, std, add=True, noise_type=NOISE_TYPE):
        """Generates a noise signal.

        Parameters
        ----------
        axis : array like
            Array for the axis.
        std : float
            Standard deviation of the noise.
        add : bool, optional
            Whether the noise should be additive (with mean 0) or multiplicative (with mean 1), by default True.
        noise_type : {"gaussian"}, optional
            The type of noise to use, by default `NOISE_TYPE` defined in the config file (gaussian).
        """
        self.methods = {
            "gaussian": NOISE._generate_gaussian,
        }
        # if noise_type == "gaussian":
        #     super().__init__(axis, NOISE._generate_gaussian(axis, std, add))
        super().__init__(axis, self.methods[noise_type](axis, std, add))

    @staticmethod
    def _generate_gaussian(axis, std, add):
        mean = 0 if add else 1
        return np.random.normal(mean, std, len(axis))


class HEAVISIDE(Signal1):
    """Heaviside step function centered at a certain point."""

    def __init__(self, axis, point=0, inverted=False):
        """Creates a Heaviside step function.

        Parameters
        ----------
        axis : array like
            Array for the axis.
        point : float, optional
            Point to center the signal around (e.g if `point == 0` then the function would change values at 0), by
            default 0.
        inverted : bool, optional
            Whether to invert the signal or not. If true, the signal would be 1 and switch to 0 after the point. By
            default False.
        """
        if inverted:
            values = list(map(lambda x: int(x <= point), axis))
        else:
            values = list(map(lambda x: int(x >= point), axis))
        super().__init__(axis, values)


class IMPULSE(Signal1):
    """Discrete impulse/delta function centered at 0."""

    def __init__(self, axis, value=1.0):
        """Creates a discrete impulse centered at 0, with value `value`.

        Parameters
        ----------
        axis : array like
            Array for the axis.
        value : float, optional
            Value for the impulse to take at 0, by default 1.0
        """
        super().__init__(axis, value * self._init_helper(axis))

    def _init_helper(self, axis):
        values = []
        for i, t in enumerate(axis):
            if t < 0.0:
                values.append(0.0)
            else:
                values.append(1.0 if t == 0 or axis[i - 1] < 0 else 0.0)
        return np.array(values)


class CONSTANT(Signal1):
    """Constant signal."""

    def __init__(self, axis, value=1.0):
        """Creates a constant signal.

        Parameters
        ----------
        axis : array like
            Array for the axis.
        value : float or complex, optional
            Value to use for the constant, by default 1.0.
        """
        super().__init__(axis, [value for t in axis])



import numpy as np
import pyqtgraph as pg
from PyQt5 import QtWidgets
import sys

from ..api import GuiInterface


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self, *args, **kwargs) -> None:
        super(MainWindow, self).__init__(*args, **kwargs)
        self.gui = GuiInterface()
        self.blocksize = 400
        # self.r1 = 0
        # self.r1 = self.blocksize // 2
        # self.r2 = self.blocksize // 2
        # self.r2 = self.blocksize
        self.values = np.zeros((1, 1))

        self.fig = pg.image(self.values)
        self.fig.setColorMap(pg.colormap.get("plasma"))

        self.send_start_request()

        self.timer = pg.Qt.QtCore.QTimer()
        self.timer.setInterval(10)
        self.timer.timeout.connect(self.update_plot_data)
        self.timer.start()

    def image(self, data):
        return self.graphWidget.image(data)

    def update_plot_data(self):
        # self.values = self.values[:, 1:]

        # max_time = 2

        new_values = self.send_fetch_request()
        # if new_values.ax0_span() > max_time:
        #     end_time = new_values.ax0[-1]
        #     new_values = new_values.get_ax0(end_time - max_time)

        self.fig.setImage(new_values.values)
        # self.values = np.vstack((self.values.T, new_col)).T

        # self.fig.setImage(self.values.T)

    def send_start_request(self):
        self.gui.make_request({
            "request_type": "start",
            "source": "microphone",
        })

    def send_fetch_request(self):
        return self.gui.make_request({
            "request_type": "spectrogram",
            "source": "microphone",
            "blocksize": self.blocksize,
            "max_time": 2,
        })

    def send_stop_request(self):
        self.gui.make_request({
            "request_type": "stop",
            "source": "microphone",
        })


def main():
    app = QtWidgets.QApplication(sys.argv)

    main = MainWindow()
    main.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



import numpy as np

from .sgn import Signal1


def w_rectangular(samp_time):
    epsilon = 0.01
    axis = [-samp_time / 2 - epsilon, -samp_time /
            2, samp_time / 2, samp_time / 2 + epsilon]
    values = [0, 1, 1, 0]
    return Signal1(axis, values)


def w_gaussian(samp_time, resolution=100):
    axis = np.linspace(-samp_time / 2, samp_time / 2, resolution)
    sigma = 3 * samp_time
    values = np.exp(-(axis ** 2) / (2 * sigma ** 2)) / \
        np.sqrt(2 * np.pi * sigma ** 2)
    return Signal1(axis, values)



"""
===================
Signal manipulation
===================

This subpackage gives the basic tools needed to create, import and
export signals of different dimensions.
"""
from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np
import bisect
import operator
import abc
from tqdm import tqdm
from numbers import Number, Real
from copy import deepcopy
from multipledispatch import dispatch

from ..exceptions import DimensionError
from ..config import (CONVOLUTION_METHOD, INTERP1_METHOD,
                      INTERP2_METHOD, CROSS_CORRELATION_METHOD,
                      KERNEL_OOB)
from .. import math_lib
from .handlers import (handler_csv, handler_json, handler_wav,
                       handler_img)


class Signal(abc.ABC):
    """Abstract class representing a signal object of arbitrary dimensions."""

    @abc.abstractmethod
    def __getitem__(self, key):
        pass

    @abc.abstractmethod
    def __add__(self, signal):
        pass

    @abc.abstractmethod
    def __sub__(self, signal):
        pass

    @abc.abstractmethod
    def __mul__(self, signal):
        pass

    @abc.abstractmethod
    def __truediv__(self, signal):
        pass

    @abc.abstractmethod
    def __eq__(self, signal):
        pass

    @abc.abstractmethod
    def __str__(self):
        pass

    @abc.abstractmethod
    def __abs__(self):
        pass

    @abc.abstractclassmethod
    def from_function(cls, axis, func, *args, **kwargs):
        """Creates a signal from an axis and a function.

        The function is applied to each element in the axis, so
        if the function f(x) = x**2 is given as a parameter to the axis
        [1, 2, 3, 4], the values would be [1, 4, 9, 16].

        Parameters
        ----------
        axis : np.ndarray
            Axis on which the function is mapped.
        func : function
            Function to map to the axis.
        """
        pass

    @abc.abstractclassmethod
    def from_file(cls, filename: str):
        """Creates a signal from a file.

        Parameters
        ----------
        filename : str
            File to read the data from.
        """
        pass

    @abc.abstractmethod
    def interpolate(self, value, method):
        """Interpolates the current values to obtain a new value."""
        pass

    @abc.abstractmethod
    def unpack(self):
        """Unpacks the signal into arrays. If used for its intended
        purpose, should be unpacked with *.
        """
        pass

    @abc.abstractmethod
    def apply_function(self, func, *args, **kwargs):
        """Applies a function to the values of the signal.

        Parameters
        ----------
        func : function
            Function to apply to the signal.
        """
        pass

    @abc.abstractmethod
    def export_to_file(self, filename: str):
        """Exports the signal values to a file.

        Parameters
        ----------
        filename : str
            File to export the data to.
        """
        pass

    @abc.abstractmethod
    def is_valid(self):
        """Checks whether the dimensions of the signal are valid."""
        pass

    def clone(self):
        """Makes a copy of this signal."""
        return deepcopy(self)

    def psd(self) -> Signal:
        """Generates the PSD (Power Spectral Density) of the signal.

        Returns
        -------
        Signal
            Signal representing the PSD.
        """
        copy = self.clone()
        copy.values = (copy.values.conjugate() * copy.values).real
        return copy

    def abs(self) -> Signal:
        """Takes the absolute value of the values."""
        return self.__abs__()

    def real_part(self) -> Signal:
        """Takes the real part of the values."""
        copy = self.clone()
        copy.values = np.real(copy.values)
        return copy

    def imag_part(self) -> Signal:
        """Takes the imaginary part of the values."""
        copy = self.clone()
        copy.values = np.imag(copy.values)
        return copy

    def conjugate(self) -> Signal:
        """Takes the conjugate of the values."""
        copy = self.clone()
        copy.values = copy.values.conjugate()
        return copy

    def shape(self) -> tuple:
        """Gets the shape of this signal."""
        return np.shape(self.values)

########################################################################################################################
# |||||||||||||||||||||||||||||||||||||||||||||||| Signal1 ||||||||||||||||||||||||||||||||||||||||||||||||||||||||||| #
########################################################################################################################


class Signal1(Signal):
    """Class representing a one dimensional signal."""
    handlers = {
        "csv": handler_csv,
        "json": handler_json,
        "wav": handler_wav,
    }

    def __init__(self, axis: np.ndarray, values: np.ndarray):
        """Creates a signal from an independent axis and a values list.

        Parameters
        ----------
        axis : array_like
            List of elements representing the independent variable
            (usually time).
        values : array_like
            List of elements representing the dependent variable for
            each axis element.

        Raises
        ------
        DimensionError
            Raises this when the dimensions of `axis` and `values`
            don't match each other.
        """
        if len(axis) != len(values):
            raise DimensionError(
                "The dimensions of the values do not match.", len(values), len(axis))
        self.axis = np.array(axis)
        self.values = np.array(values)

    def __getitem__(self, key):
        return self.values[key]

    @dispatch(slice)
    def __call__(self, key):
        # Slices the indices based on the given key, then intersects
        # them to get all the indices
        indices1 = np.where(
            key.start <= self.axis if key.start else self.axis)
        indices2 = np.where(
            self.axis <= key.stop if key.stop else self.axis)
        indices = np.intersect1d(indices1, indices2)
        return [self.values[i] for i in indices]

    def __call__(self, key, inter_method=INTERP1_METHOD):
        return self.interpolate(key, inter_method)[2]

    def __radd__(self, num):
        return self.__add__(num)

    @dispatch(Number)
    def __add__(self, value):
        return Signal1(self.axis, self.values + value)

    @dispatch(object)
    def __add__(self, signal):
        return Signal1(*self._do_bin_operation(signal, operator.add))

    def __rsub__(self, num):
        return num + self * -1

    @dispatch(Number)
    def __sub__(self, value):
        return Signal1(self.axis, self.values - value)

    @dispatch(object)
    def __sub__(self, signal):
        return Signal1(*self._do_bin_operation(signal, operator.sub))

    def __rmul__(self, num):
        return self.__mul__(num)

    @dispatch(Number)
    def __mul__(self, value):
        return Signal1(self.axis, self.values * value)

    @dispatch(object)
    def __mul__(self, signal):
        return Signal1(*self._do_bin_operation(signal, operator.mul))

    def __rtruediv__(self, num):
        return Signal1(self.axis, num / self.values)

    @dispatch(Number)
    def __truediv__(self, value):
        return Signal1(self.axis, self.values / value)

    @dispatch(object)
    def __truediv__(self, signal):
        return Signal1(*self._do_bin_operation(signal, operator.truediv))

    def __eq__(self, signal):
        return (
            np.array_equal(self.axis, signal.axis)
            and np.array_equal(self.values, signal.values)
        )

    def __str__(self):
        return f"{self.axis}\n{self.values}"

    def __abs__(self):
        return Signal1(self.axis, list(map(operator.abs, self.values)))

    def __len__(self):
        return len(self.axis)

    def _do_bin_operation(self, signal, operation, inter_method=INTERP1_METHOD, debug=False):
        # Joins the axes of both signals
        axis_list = np.union1d(self.axis, signal.axis)
        # axis_list.sort()

        new_values = np.array([])
        iterable = tqdm(
            axis_list, "Applying operation") if debug else axis_list
        for t in iterable:
            # Interpolates the values
            y1 = self(t, inter_method)
            y2 = signal(t, inter_method)
            # Operates using the interpolated values
            new_values = np.append(new_values, operation(y1, y2))
        return axis_list, new_values

    @classmethod
    def from_function(cls, axis: np.ndarray, func, *args, **kwargs):
        """Creates a signal from an axis list and a function.

        The function is applied to each element in the axis, so if the
        function f(x) = x**2 is given as a parameter to the axis
        [1, 2, 3, 4], the values should be [1, 4, 9, 16].

        Parameters
        ----------
        axis : array_like
            List of elements representing the independent variable
            (usually time).
        func : function
            Function to apply to each element.
        """
        return cls(axis, func(np.array(axis), *args, **kwargs))

    @classmethod
    def from_file(cls, filename: str, *args, **kwargs):
        """Creates a signal from a file.

        Parameters
        ----------
        filename : str
            Name of the file, including its path.

        Returns
        -------
        Signal1
            Signal after being read.

        Raises
        ------
        ValueError
            [description]
        """
        extension = filename.split(".")[-1]
        if extension == filename:
            raise ValueError()
        return cls(*Signal1.handlers[extension].import_signal1(
            filename, *args, **kwargs
        ))

    @classmethod
    def from_freq(cls, values: np.ndarray, sf=1, sp=0):
        """Creates a signal from a values list and a sampling frequency.

        Parameters
        ----------
        values : array_like
            List of elements representing the dependent variable for
            each axis element.
        sf : real number, optional
            Sampling frequency used to create the axis, by default 1.
        sp : real number, optional
            Starting point for the axis, by default 0.
        """
        samp_period = 1 / sf
        vals = np.array(values)
        axis = samp_period * np.arange(len(vals)) - sp
        return cls(axis, vals)

    @dispatch(Number, str)
    def add(self, value, method=INTERP1_METHOD, *args, **kwargs):
        """Adds this signal with another value."""
        return Signal1(self.axis, self.values + value)

    @dispatch(object, str)
    def add(self, signal, method=INTERP1_METHOD):
        """Adds this signal with another value."""
        return Signal1(*self._do_bin_operation(signal, operator.add, method))

    @dispatch(Number, str)
    def sub(self, value, method=INTERP1_METHOD):
        """Subtracts this signal with another value."""
        return Signal1(self.axis, self.values - value)

    @dispatch(object, str)
    def sub(self, signal, method=INTERP1_METHOD, *args, **kwargs):
        """Subtracts this signal with another value."""
        return Signal1(*self._do_bin_operation(signal, operator.sub, method, *args, **kwargs))

    @dispatch(Number, str)
    def mul(self, value, method=INTERP1_METHOD, *args, **kwargs):
        """Multiplies this signal with another value."""
        return Signal1(self.axis, self.values * value)

    @dispatch(object, str)
    def mul(self, signal, method=INTERP1_METHOD, *args, **kwargs):
        """Multiplies this signal with another value."""
        return Signal1(*self._do_bin_operation(signal, operator.mul, method, *args, **kwargs))

    @dispatch(Number, str)
    def div(self, value, method=INTERP1_METHOD, *args, **kwargs):
        """Divides this signal with another value."""
        return Signal1(self.axis, self.values / value)

    @dispatch(object, str)
    def div(self, signal, method=INTERP1_METHOD, *args, **kwargs):
        """Divides this signal with another value."""
        return Signal1(*self._do_bin_operation(signal, operator.truediv, method, *args, **kwargs))

    def sampling_freq(self) -> float:
        """Calculates the sampling frequency in hertz, assuming it is constant."""
        sf = 1 / (self.axis[1] - self.axis[0])
        return sf if sf > 0 else 0

    def interpolate_list(self, elements: list, method=INTERP1_METHOD):
        """Interpolates the current values to obtain new ones.

        Parameters
        ----------
        elements : list
            List of elements to interpolate.
        method : {"linear", "sinc"}, optional
            Method used for the interpolation, by default INTERP1_METHOD.

        Returns
        -------
        copy : Signal1
            Copy of the signal with the new values interpolated.
        """
        copy = self.clone()
        for t in elements:
            copy, _, _ = copy.interpolate(t, method)
        return copy

    def interpolate(self, element, method=INTERP1_METHOD):
        """Interpolates the current values to obtain a new value.

        Parameters
        ----------
        element : float
            Element to apply the interpolation to.
        method : {"linear", "sinc"}, optional
            Method used for the interpolation, by default INTERP1_METHOD.
        Returns
        -------
        copy : Signal1
            Copy of the signal with the new value interpolated.
        index : int
            Index of the interpolated value.
        new_value : float
            Value of the interpolated value.
        """
        methods = {
            "linear": self._linear_interp,
            "sinc": self._sinc_interp,
        }
        copy = self.clone()

        if element not in self.axis:
            return methods[method](element)
        else:
            index = bisect.bisect(copy.axis, element) - 1
            return copy, index, self[index]

    def _linear_interp(self, element):
        copy = self.clone()
        new_index = bisect.bisect(self.axis, element)
        copy.axis = np.insert(copy.axis, new_index, element)
        copy.values = np.insert(copy.values, new_index, 0)

        ta = copy.axis[new_index - 1]
        xa = copy.values[new_index - 1]
        try:
            tb = copy.axis[new_index + 1]
            xb = copy.values[new_index + 1]
        except IndexError:
            # This code is reached if the program tries to
            # interpolate points out of the range. In this case,
            # it simply interpolates using the last value. For
            # `xb` we take the element -2 because, if this code
            # is reached, a 0 was added in the last value
            tb = copy.axis[-1]
            xb = copy.values[-2]

        # Linearly interpolates
        new_value = xa + (xb - xa) * (element - ta) / (tb - ta)
        copy.values[new_index] = new_value
        return copy, new_index, new_value

    def _sinc_interp(self, element):
        copy = self.clone()
        new_index = bisect.bisect(self.axis, element)
        copy.axis = np.insert(copy.axis, new_index, element)
        copy.values = np.insert(copy.values, new_index, 0)

        fs = copy.sampling_freq()
        result = 0
        for t, x in zip(*self.unpack()):
            result += x * np.sinc(fs * (element - t))

        copy.values[new_index] = result
        return copy, new_index, result

    def unpack(self):
        """Unpacks the signal into two arrays. If used for its
        intended purpose, should be unpacked with *.
        """
        return self.axis, self.values

    def span(self) -> float:
        """Gets the span of the signal"""
        return self.axis[-1] - self.axis[0]

    def half(self, first=True):
        """Gets half of the signal"""
        half_span = int(len(self) / 2)
        if first:
            return Signal1(self.axis[:half_span], self.values[:half_span])
        else:
            return Signal1(self.axis[half_span:], self.values[half_span:])
        # return self[:int(self.span() / 2)] * 2 if first else self[int(self.span() / 2):] * 2

    def rect_smooth(self, factor: int) -> Signal1:
        """Directly applies a rectangular smoothing to the signal.

        With this method the edges of the signal look a bit rough.

        Parameters
        ----------
        factor : int (odd)
            Smoothing factor.

        Returns
        -------
        Signal1
            Smooth signal.
        """
        copy = self.clone()
        if factor % 2 != 1 or factor <= 1:
            raise ValueError("The smoothing factor must be an odd number.")
        shift = int((factor - 1) / 2)
        self_len = len(copy)
        new_values = copy.values[0:1]               # Copies the first element

        # Smooths the first elements with the only possible elements
        for n in range(1, shift):
            arr = copy.values[0:2 * n + 1]
            new_values = np.append(new_values, arr.sum() / (2 * n + 1))

        # Smooths the other elements using the given factor
        for n in range(shift, self_len - shift):
            arr = copy.values[n - shift:n + shift + 1]
            new_values = np.append(new_values, arr.sum() / factor)

        # Smooths the last elements adapting the smoothing factor
        for n in range(self_len - shift, self_len):
            new_shift = self_len - n - 1
            arr = copy.values[n - new_shift:self_len]
            new_values = np.append(new_values, arr.sum() / (2 * new_shift + 1))

        assert self_len == len(
            new_values), "There was an error during the smoothing."
        copy.values = new_values
        return copy

    def apply_function(self, func, *args, **kwargs) -> Signal1:
        """Applies a function to the values of the signal.

        Parameters
        ----------
        func : function
            Function to apply to the signal.

        Returns
        -------
        Signal1
            Modified signal.
        """
        copy = self.clone()
        copy.values = np.array([func(x, *args, **kwargs) for x in copy.values])
        return copy

    def apply_function_tuple(self, func, *args, **kwargs) -> Signal1:
        """Applies a function to both the axis and values of the signal.

        Parameters
        ----------
        func : function
            Function to apply to the signal.

        Returns
        -------
        Signal1
            Modified signal.
        """
        copy = self.clone()
        copy.values = np.array([func(t, x, *args, **kwargs)
                                for t, x in zip(copy.axis, copy.values)])
        return copy

    def convolute(self, signal1: Signal1, method=CONVOLUTION_METHOD) -> Signal1:
        """Convolute this signal with another.

        Parameters
        ----------
        signal1 : Signal1
            Signal to convolute with.
        method : {"fft", "direct"}, optional
            Method utilized to calculate the convolution, by
            default CONVOLUTION_METHOD.

        Returns
        -------
        Signal1
            Convoluted signal.
        """
        return math_lib.convolution(self, signal1, method)

    def cross_correlate(self, signal1: Signal1,
                        method=CROSS_CORRELATION_METHOD) -> Signal1:
        """Cross-correlates this signal with another.

        Parameters
        ----------
        signal1 : Signal1
            Signal to cross-correlate with.
        method : {"direct"}, optional
            Method utilized to calculate the cross-correlation, by
            default CROSS_CORRELATION_METHOD.

        Returns
        -------
        Signal1
            Cross-correlated signal.
        """
        return math_lib.cross_correlation(self, signal1, method)

    def auto_correlate(self, method=CROSS_CORRELATION_METHOD) -> Signal1:
        """Auto-correlates this signal.

        Parameters
        ----------
        method : {"direct"}, optional
            Method utilized to calculate the auto-correlation, by
            default CROSS_CORRELATION_METHOD.

        Returns
        -------
        Signal1
            Auto-correlated signal.
        """
        return math_lib.cross_correlation(self, self, method)

    def shift(self, value) -> Signal1:
        """Shifts the axis by `value`."""
        copy = self.clone()
        copy.axis += value
        return copy

    def export_to_file(self, filename: str, *args, **kwargs):
        """Exports the one dimensional signal to the given file.

        Parameters
        ----------
        filename : str
            String corrresponding to the file.

        Raises
        ------
        ValueError
            If the filename is empty (e.g trying to export to the file ".csv").
        """
        extension = filename.split(".")[-1]
        if extension == filename:
            raise ValueError()
        Signal1.handlers[extension].export_signal1(
            filename, self, *args, **kwargs)

    def apply_window(self, window: Signal1, center: Real,
                     interp_method=INTERP1_METHOD, *args,
                     **kwargs) -> Signal1:
        """Applies a window function to the signal.

        For this implementation it is assumed that the window function
        is zero outside of its given axis.

        Parameters
        ----------
        window : Signal1
            Window signal.
        center : Real
            Center point where the window is applied.
        interp_method : string, optional
            Method used for the interpolation, by default INTERP1_METHOD

        Returns
        -------
        Signal1
            Signal after applying the window.
        """
        w_span = window.span()
        copy = self.clone()

        # Divides the signals into the three relevant parts
        l_signal = copy.get(stop=center - w_span / 2)
        c_signal = copy.get(center - w_span / 2, center + w_span / 2)
        r_signal = copy.get(start=center + w_span / 2)

        # We assume the signals are zero outside of their specified range
        l_signal.values = np.zeros(len(l_signal))
        r_signal.values = np.zeros(len(r_signal))

        # We apply the window to the center part
        c_signal = c_signal.mul(window.clone().shift(
            center), interp_method, *args, **kwargs)

        return l_signal.concatenate(c_signal, r_signal)

    def get(self, start=None, stop=None) -> Signal1:
        """Gets a portion of the signal.

        Parameters
        ----------
        start : float, optional
            Starting point, by default the first point.
        stop : float, optional
            Stopping point, by default the last point.

        Returns
        -------
        Signal1
            The cut signal.
        """
        copy = self.clone()
        if start is None:
            start_index = 0
        else:
            start_index = bisect.bisect(copy.axis, start)

        if stop is None:
            stop_index = -1
        else:
            stop_index = bisect.bisect(copy.axis, stop)

        copy.axis, copy.values = (copy.axis[start_index:stop_index],
                                  copy.values[start_index:stop_index])
        return copy

    def concatenate(self, *signals) -> Signal1:
        """Concatenates this signal with others."""
        copy = self.clone()
        s_axis, s_values = ((sign.axis for sign in signals),
                            (sign.values for sign in signals))
        return Signal1(np.concatenate((copy.axis, *s_axis)),
                       np.concatenate((copy.values, *s_values)))

    def is_valid(self):
        return self.axis.shape == self.values.shape

########################################################################################################################
# |||||||||||||||||||||||||||||||||||||||||||||||| Signal2 ||||||||||||||||||||||||||||||||||||||||||||||||||||||||||| #
########################################################################################################################


class Signal2(Signal):
    """Class representing a two dimensional signal."""
    handlers = {
        "jpeg": handler_img,
        "jpg": handler_img,
        "png": handler_img,
    }

    def __init__(self, ax0: np.ndarray, ax1: np.ndarray,
                 values: np.ndarray):
        """Creates a two dimensional signal by giving two axes and a
        matrix.

        Each axis corresponds to one of the dimensions, where
        `ax0` indexes the rows of `values`, while `ax1` indexes its
        columns.

        Parameters
        ----------
        ax0 : array_like
            First axis, which indexes the rows of `values`.
        ax1 : array_like
            Second axis, which indexes the columns of `values`.
        values : two-dimensional array_like
            Matrix that indicates the values of the signal for every
            point.

        Raises
        ------
        DimensionError
            Raises this when the shape of `values` doesn't match the
            sizes of `ax0` and `ax1`.

        Example
        -------
        Creating the following object
        >>> ax0 = [1, 2, 3]
        >>> ax1 = [2, 4, 6]
        >>> vals = [
        >>>     [1, 2, 3],
        >>>     [2, 4, 6],
        >>>     [3, 6, 9]
        >>> ]
        >>> signal = Signal2(ax0, ax1, vals)
        can be understood as the following plot
          ax1
           |
         6 |  3  6  9
           |
         4 |  2  4  6
           |
         2 |  1  2  3
           |
         0 |--------- ax0
           0  1  2  3
        """
        if np.shape(values) != (len(ax0), len(ax1)):
            raise DimensionError("The dimensions of the values do not match.", np.shape(
                values), (len(ax0), len(ax1)))
        self.ax0 = np.array(ax0)
        self.ax1 = np.array(ax1)
        self.values = np.array(values)

    def __getitem__(self, key):
        return self.values[key]

    @dispatch(Real, Real)
    def __call__(self, key_x, key_y):
        return self.interpolate(key_x, key_y)[2]

    @dispatch(Real, Real, str)
    def __call__(self, key_x, key_y, interp_method=INTERP2_METHOD):
        return self.interpolate(key_x, key_y, method=interp_method)[2]

    def __radd__(self, num):
        return self.__add__(num)

    # TODO: Fix the REALLY slow speed when operating signals
    @dispatch(Number)
    def __add__(self, value):
        return Signal2(self.ax0, self.ax1, self.values + value)

    @dispatch(object)
    def __add__(self, signal):
        return Signal2(*self._do_bin_operation(signal, operator.add))

    def __rsub__(self, num):
        return num + self * -1

    @dispatch(Number)
    def __sub__(self, value):
        return Signal2(self.ax0, self.ax1, self.values - value)

    @dispatch(object)
    def __sub__(self, signal):
        return Signal2(*self._do_bin_operation(signal, operator.sub))

    def __rmul__(self, num):
        return self.__mul__(num)

    @dispatch(Number)
    def __mul__(self, value):
        return Signal2(self.ax0, self.ax1, self.values * value)

    @dispatch(object)
    def __mul__(self, signal):
        return Signal2(*self._do_bin_operation(signal, operator.mul))

    def __rtruediv__(self, num):
        return Signal2(self.ax0, self.ax1, num / self.values)

    @dispatch(Number)
    def __truediv__(self, value):
        return Signal2(self.ax0, self.ax1, self.values / value)

    @dispatch(object)
    def __truediv__(self, signal):
        return Signal2(*self._do_bin_operation(signal, operator.truediv))

    def __eq__(self, signal):
        return (
            np.array_equal(self.ax0, signal.ax0)
            and np.array_equal(self.ax1, signal.ax1)
            and np.array_equal(self.values, signal.values)
        )

    def __str__(self):
        return f"{self.ax0}\n{self.ax1}\n{self.values}"

    def __abs__(self):
        copy = self.clone()
        copy.values = abs(copy.values)
        return copy

    def _do_bin_operation(self, signal, operation):
        # Joins the axes of both signals
        new_ax0 = np.union1d(self.ax0, signal.ax0)
        new_ax1 = np.union1d(self.ax1, signal.ax1)
        new_ax0.sort()
        new_ax1.sort()

        new_values = new_ax1.copy()
        for x in new_ax0:
            row = np.array([])
            for y in new_ax1:
                # Interpolates the values
                val1 = self(x, y)
                val2 = signal(x, y)
                # Operates using the interpolated values
                row = np.append(row, operation(val1, val2))
            new_values = np.vstack((new_values, row))
        return new_ax0, new_ax1, new_values

    @classmethod
    def from_function(cls, ax0, ax1, func, *args, **kwargs):
        """Creates a signal from two axes and a function.

        The function is applied to each element in the axis, so
        if the function `f(x, y) = x**2 + y**2` is given as a parameter to
        the axes `[1, 2, 3]` and `[-1, -2, -3]`, the values would be the
        matrix `[[2, 5, 10], [5, 8, 13], [10, 13, 18]]`.

        Parameters
        ----------
        ax0 : np.ndarray
            First on which the function is mapped.
        ax1 : np.ndarray
            Second on which the function is mapped.
        func : function
            Function to map to the axes.
        """
        values = np.array([[func(x, y, *args, **kwargs)
                          for x in ax0] for y in ax1])
        return cls(ax0, ax1, values)

    @classmethod
    def from_file(cls, filename: str, *args, **kwargs):
        """Creates a signal from a file. If the file is an image with
        an RGB channel, using `channel` you can specify which channel
        to read from, or the method used to handle them.

        Parameters
        ----------
        filename : str
            File to read the data from.
        """
        extension = filename.split(".")[-1]
        if extension == filename:
            raise ValueError()
        return cls(*Signal2.handlers[extension].import_signal2(filename, *args, **kwargs))

    @classmethod
    def from_freq(cls, values: np.ndarray, sf_ax0=1, sf_ax1=1, sp_ax0=0, sp_ax1=0):
        """Creates a two dimensional signal by giving a values matrix
        and a frequency for each axis.

        Each axis corresponds to one of the dimensions, where
        `ax0` indexes the rows of `values`, while `ax1` indexes its
        columns.

        Parameters
        ----------
        values : two-dimensional array_like
            Matrix that indicates the values of the signal for every
            point.
        sf_ax0 : float, optional
            Sampling frequency of the first axis, by default 1.
        sf_ax1 : float, optional
            Sampling frequency of the second axis, by default 1.
        sp_ax0 : float, optional
            Starting point for the first axis, by default 0.
        sp_ax1 : float, optional
            Starting point for the second axis, by default 0.
        """
        ax0_samp_period = 1 / sf_ax0
        ax1_samp_period = 1 / sf_ax1
        vals = np.array(values)
        val_shape = np.shape(values)

        ax0 = np.arange(val_shape[0]) * ax0_samp_period - sp_ax0
        ax1 = np.arange(val_shape[1]) * ax1_samp_period - sp_ax1
        return cls(ax0, ax1, vals)

    @dispatch(Real, Real)
    def interpolate(self, val0, val1):
        """Interpolates the current values to obtain a new value."""
        return self._interpolate(val0, val1)

    @dispatch(Real, Real, str)
    def interpolate(self, val0, val1, method=INTERP2_METHOD):
        """Interpolates the current values to obtain a new value."""
        return self._interpolate(val0, val1, method)

    def _interpolate(self, val0, val1, method=INTERP2_METHOD):
        methods = {
            "bilinear": self._bilinear_interp,
        }
        copy = self.clone()

        if val0 not in self.ax0 and val1 not in self.ax1:
            return methods[method](val0, val1)
        else:
            ind0 = bisect.bisect(copy.ax0, val0) - 1
            ind1 = bisect.bisect(copy.ax1, val1) - 1
            return copy, (ind0, ind1), self[ind0, ind1]

    def _bilinear_interp(self, val0, val1):
        copy = self.clone()
        new_ind0 = bisect.bisect(self.ax0, val0)
        new_ind1 = bisect.bisect(self.ax1, val1)
        copy.ax0 = np.insert(copy.ax0, new_ind0, val0)
        copy.ax1 = np.insert(copy.ax1, new_ind1, val1)
        copy.values = np.insert(
            np.insert(copy.values, new_ind0, 0, 0), new_ind1, 0, 1)
        val_shape = copy.values.shape

        copy.values[new_ind0, new_ind1] = self._bilinear_interp_point(
            copy, new_ind0, new_ind1, val0, val1)

        for i in range(new_ind0):
            copy.values[i, new_ind1] = self._interp_side_neighbors(
                copy, i, new_ind1, copy.ax0[i], val1, axis=1)
        for i in range(new_ind1 + 1, val_shape[0]):
            copy.values[i, new_ind1] = self._interp_side_neighbors(
                copy, i, new_ind1, copy.ax0[i], val1, axis=1)

        for j in range(new_ind1):
            copy.values[new_ind0, j] = self._interp_side_neighbors(
                copy, new_ind0, j, val0, copy.ax1[j], axis=0)
        for j in range(new_ind1 + 1, val_shape[1]):
            copy.values[new_ind0, j] = self._interp_side_neighbors(
                copy, new_ind0, j, val0, copy.ax1[j], axis=0)

        return copy, (new_ind0, new_ind1), copy[new_ind0, new_ind1]

    def _bilinear_interp_point(self, copy, new_ind0, new_ind1, val0, val1):
        # We calculate all the required indices
        x0_error = False
        x1_error = False
        y0_error = False
        y1_error = False

        try:
            x0 = copy.ax0[new_ind0 - 1]
        except IndexError:
            x0_error = True
            x0 = copy.ax0[0]
        try:
            y0 = copy.ax1[new_ind1 - 1]
        except IndexError:
            y0_error = True
            y0 = copy.ax1[0]

        try:
            x1 = copy.ax0[new_ind0 + 1]
        except IndexError:
            x1_error = True
            x1 = copy.ax0[-1]
        try:
            y1 = copy.ax1[new_ind1 + 1]
        except IndexError:
            y1_error = True
            y1 = copy.ax1[-1]

        ind0 = 1 if x0_error else new_ind0 - 1
        ind1 = 1 if y0_error else new_ind1 - 1
        f00 = copy.values[ind0, ind1]

        ind0 = 1 if x0_error else new_ind0 - 1
        ind1 = -2 if y1_error else new_ind1 - 1
        f01 = copy.values[ind0, ind1]

        ind0 = -2 if x1_error else new_ind0 - 1
        ind1 = 1 if y0_error else new_ind1 - 1
        f10 = copy.values[ind0, ind1]

        ind0 = -2 if x1_error else new_ind0 - 1
        ind1 = -2 if y1_error else new_ind1 - 1
        f11 = copy.values[ind0, ind1]

        # With them, we interpolate in the x direction
        fx0 = f00 * (x1 - val0) / (x1 - x0) + f10 * (val0 - x0) / (x1 - x0)
        fx1 = f01 * (x1 - val0) / (x1 - x0) + f11 * (val0 - x0) / (x1 - x0)

        # Now we interpolate in the y direction
        return fx0 * (y1 - val1) / (y1 - y0) + fx1 * (val1 - y0) / (y1 - y0)

    def _interp_side_neighbors(self, copy, new_ind0, new_ind1, val0, val1, axis=0):
        if axis == 0:
            x0_error = False
            x1_error = False

            try:
                x0 = copy.ax0[new_ind0 - 1]
            except IndexError:
                x0_error = True
                x0 = copy.ax0[0]

            try:
                x1 = copy.ax0[new_ind0 + 1]
            except IndexError:
                x1_error = True
                x1 = copy.ax0[-1]

            ind0 = 1 if x0_error else new_ind0 - 1
            f0 = copy.values[ind0, new_ind1]

            ind0 = -2 if x1_error else new_ind0 + 1
            f1 = copy.values[ind0, new_ind1]

            # Linearly interpolates
            return f0 + (f1 - f0) * (val0 - x0) / (x1 - x0)
        elif axis == 1:
            y0_error = False
            y1_error = False

            try:
                y0 = copy.ax1[new_ind1 - 1]
            except IndexError:
                y0_error = True
                y0 = copy.ax1[0]

            try:
                y1 = copy.ax1[new_ind1 + 1]
            except IndexError:
                y1_error = True
                y1 = copy.ax0[-1]

            ind1 = 1 if y0_error else new_ind1 - 1
            f0 = copy.values[new_ind0, ind1]

            ind1 = -2 if y1_error else new_ind1 + 1
            f1 = copy.values[new_ind0, ind1]

            # Linearly interpolates
            return f0 + (f1 - f0) * (val1 - y0) / (y1 - y0)

    def unpack(self):
        """Unpacks the signal into three arrays. If used for its
        intended purpose, should be unpacked with *.
        """
        return self.ax0, self.ax1, self.values

    def apply_function(self, func, *args, **kwargs):
        """Applies a function to the values of the signal.

        Parameters
        ----------
        func : function
            Function to apply to the signal.
        """
        copy = self.clone()
        copy.values = np.array([func(x, *args, **kwargs) for x in copy.values])
        return copy

    def export_to_file(self, filename: str, *args, **kwargs):
        """Exports the one dimensional signal to the given file.

        Parameters
        ----------
        filename : str
            String corrresponding to the file.

        Raises
        ------
        ValueError
            If the filename is empty (e.g trying to export to the file ".csv").
        """
        extension = filename.split(".")[-1]
        if extension == filename:
            raise ValueError()
        Signal2.handlers[extension].export_signal2(
            filename, self, *args, **kwargs)

    def ax0_sampling_freq(self) -> float:
        """Calculates the sampling frequency of `ax0` in hertz, assuming
        it is constant.
        """
        sf = 1 / (self.ax0[1] - self.ax0[0])
        return sf if sf > 0 else 0

    def ax1_sampling_freq(self) -> float:
        """Calculates the sampling frequency of `ax1` in hertz, assuming
        it is constant.
        """
        sf = 1 / (self.ax1[1] - self.ax1[0])
        return sf if sf > 0 else 0

    def ax0_span(self) -> float:
        """Gets the span of the first axis."""
        return self.ax0[-1] - self.ax0[0]

    def ax1_span(self) -> float:
        """Gets the span of the second axis."""
        return self.ax1[-1] - self.ax1[0]

    def apply_kernel(self, kernel: np.ndarray, flip=False, oob=KERNEL_OOB) -> Signal2:
        """Applies a kernel over the signal. This process is also known
        as image convolution.

        Parameters
        ----------
        kernel : np.ndarray
            Matrix of the kernel to apply to the signal.
        flip : bool, optional
            Wheter to flip the kernel or not, by default False.
        oob : str, optional
            Specifier for how to handle values outside of the bounds of
            the signal, by default KERNEL_OOB.

        Returns
        -------
        Signal2
            Signal after applying the kernel.
        """
        return math_lib.apply_kernel(self, kernel, flip, oob)

    def transpose(self) -> Signal2:
        """Transposes the signal by interchanging `ax0` and `ax1`, and
        taking the transpose of `values`.

        Returns
        -------
        Signal2
            Transposed signal.
        """
        copy = self.clone()
        copy.ax0, copy.ax1 = copy.ax1, copy.ax0
        copy.values = copy.values.T
        assert np.shape(copy.values) == (
            len(copy.ax0), len(copy.ax1)), "Something went wrong."
        return copy

    def contourf(self):
        """Unpacks the signal in a way that the function `contourf`
        within the module `matplotlib.pyplot` can easily understand.
        If used for this purpose, should be called with *. 

        For example, if you want to plot the signal `sign`, then you
        would call
        >>> plt.contourf(*sign.contourf())

        For this, it returns both axes and the values (like the
        `unpack` method), except that the values are transposed.

        Returns
        -------
        np.ndarray, np.ndarray, np.ndarray
            Attributes `ax0`, `ax1` and `values` (this last one is
            transposed).
        """
        return self.ax0, self.ax1, self.values.T

    def imshow(self):
        """Unpacks the signal in a way that the function `imshow` within
        the module `matplotlib.pyplot` can easily understand, in such a
        way that the axes are automatically reshaped to fit the real
        axes of the signal. If used for this purpose, should be
        called with **.

        For example, if you want to plot the signal `sign`, then
        you would call
        >>> plt.imshow(**sign.imshow())

        Returns
        -------
        dict
            Dictionary with the appropiate keywords for `plt.imshow`.
        """
        copy = self.clone()
        xmin, xmax = copy.ax0[0], copy.ax0[-1]
        ymin, ymax = copy.ax1[0], copy.ax1[-1]
        return {"X": copy.values.T, "extent": [xmin, xmax, ymin, ymax]}

    def half(self, axis=1, first=False):
        """Gets half of the signal.

        Which half to take can be specified using the `axis` and `first`
        parameters. `axis` tells the program in which direction to make
        the cut (axis 0 is row-wise, and axis 1 is column-wise), while
        `first` indicates whether to take the first half or second half.

        Parameters
        ----------
        axis : int, optional
            Direction in which to make the cut, by default 1. `0` is
            row-wise, `1` is column-wise.
        first : bool, optional
            Whether to take the first or second half, by default False.

        Returns
        -------
        Signal2
            Signal cut in half.
        """
        ax_handlers = {
            0: self._half_0,
            1: self._half_1,
        }
        return ax_handlers[axis](first)

    def _half_0(self, first=False):
        copy = self.clone()
        c_shape = copy.shape()
        half_val = int(c_shape[0] / 2)
        if first:
            copy.ax0 = copy.ax0[:half_val]
            copy.values = copy.values[:half_val, :]
        else:
            copy.ax0 = copy.ax0[half_val:]
            copy.values = copy.values[half_val:, :]
        return copy

    def _half_1(self, first=False):
        copy = self.clone()
        c_shape = copy.shape()
        half_val = int(c_shape[1] / 2)
        if first:
            copy.ax1 = copy.ax1[:half_val]
            copy.values = copy.values[:, :half_val]
        else:
            copy.ax1 = copy.ax1[half_val:]
            copy.values = copy.values[:, half_val:]
        return copy

    def is_valid(self):
        return self.values.shape == (len(self.ax0), len(self.ax1))

    def get_ax0(self, start=None, stop=None) -> Signal2:
        copy = self.clone()
        if start is None:
            start_index = 0
        else:
            start_index = bisect.bisect(copy.ax0, start)

        if stop is None:
            stop_index = -1
        else:
            stop_index = bisect.bisect(copy.ax0, stop)

        copy.ax0 = copy.ax0[start_index:stop_index]
        copy.values = copy.values[start_index:stop_index, :]
        return copy

    def get_ax1(self, start=None, stop=None) -> Signal2:
        copy = self.clone()
        if start is None:
            start_index = 0
        else:
            start_index = bisect.bisect(copy.ax1, start)

        if stop is None:
            stop_index = -1
        else:
            stop_index = bisect.bisect(copy.ax1, stop)

        copy.ax1 = copy.ax1[start_index:stop_index]
        copy.values = copy.values[:, start_index:stop_index]
        return copy



from __future__ import annotations
from typing import TYPE_CHECKING
import sys
import os
import logging
import pyqtgraph as pg
import numpy as np
from multipledispatch import dispatch
from PyQt5 import QtCore, QtWidgets, QtGui

import chirper
from ..api import GuiInterface
if TYPE_CHECKING:
    from ..sgn import Signal

########################################################################################################################
# ||||||||||||||||||||||||||||||||||||||||||||||| Main container ||||||||||||||||||||||||||||||||||||||||||||||||||||| #
########################################################################################################################


class ChirperApp(QtWidgets.QMainWindow):
    """Main window for Chirper."""

    def __init__(self) -> None:
        super().__init__()
        self.title = "Test"
        self.left = 100
        self.top = 100
        self.width = 640
        self.height = 480

        self.init_ui()

    def set_stylesheet(self, filename) -> None:
        """Read a stylesheet."""
        with open(filename, "r") as stylesheet:
            self.setStyleSheet(stylesheet.read())

    def init_ui(self) -> None:
        """Initializes the main window."""
        self.set_stylesheet(os.path.join(
            chirper.BASE_DIRNAME, "gui", "styles.css"))
        self.setWindowTitle(self.title)
        self.setGeometry(self.left, self.top, self.width, self.height)
        self.make_menu_bar()

        self.main_window = ChirperMainWidget(self)
        self.setCentralWidget(self.main_window)

        self.make_permanent_status_bar()
        self.show()

    def add_entry_to_menu(self, title, bar, shortcut=None,
                          on_pressed=None) -> None:
        """Adds an entry to the top menu."""
        act = QtWidgets.QAction(title, bar)
        if shortcut:
            act.setShortcut(shortcut)
        if on_pressed:
            act.triggered.connect(on_pressed)
        bar.addAction(act)

    def make_menu_bar(self) -> None:
        """Makes the top menu bar."""
        self.bar = self.menuBar()

        self.bar_file = self.bar.addMenu("File")
        self.add_entry_to_menu("&New", self.bar_file,
                               "Ctrl+N", lambda: self.menu_act("New"))
        self.add_entry_to_menu("&Open", self.bar_file,
                               "Ctrl+O", lambda: self.menu_act("Open"))
        self.add_entry_to_menu("&Save", self.bar_file,
                               "Ctrl+S", lambda: self.menu_act("Save"))
        self.add_entry_to_menu("&Save As", self.bar_file,
                               "Ctrl+Shift+S", lambda: self.menu_act("Save As"))

        self.bar_file.addSeparator()

        self.export_menu = self.bar_file.addMenu("Export")
        self.add_entry_to_menu("&csv", self.export_menu,
                               on_pressed=lambda: self.menu_act("csv"))
        self.add_entry_to_menu("&json", self.export_menu,
                               on_pressed=lambda: self.menu_act("json"))
        self.add_entry_to_menu("&wav", self.export_menu,
                               on_pressed=lambda: self.menu_act("wav"))
        self.add_entry_to_menu("&png", self.export_menu,
                               on_pressed=lambda: self.menu_act("png"))

        self.bar_file.addSeparator()

        self.add_entry_to_menu("&Exit", self.bar_file,
                               "Ctrl+W", self.close)

        self.bar_edit = self.bar.addMenu("Edit")
        self.add_entry_to_menu("&Undo", self.bar_edit,
                               "Ctrl+Z", lambda: self.menu_act("Undo"))
        self.add_entry_to_menu("&Redo", self.bar_edit,
                               "Ctrl+Y", lambda: self.menu_act("Redo"))

        self.bar_view = self.bar.addMenu("View")
        self.add_entry_to_menu("&Zoom", self.bar_view,
                               on_pressed=lambda: self.menu_act("Zoom"))
        self.add_entry_to_menu("&Fit to content", self.bar_view,
                               on_pressed=lambda: self.menu_act("Fit to content"))

        self.bar_help = self.bar.addMenu("Help")
        self.add_entry_to_menu("&README", self.bar_help,
                               on_pressed=lambda: self.menu_act("README"))
        self.add_entry_to_menu("&Visit webpage", self.bar_help,
                               on_pressed=lambda: self.menu_act("Visit webpage"))

    def menu_act(self, btn, dur=5000) -> None:
        self.statusBar().showMessage(f"Pressed {btn}", dur)

    def make_permanent_status_bar(self) -> None:
        """Adds permanent information to the status bar."""
        status_bar_msg = f"Chirper v{chirper.__version__} - For more information, visit the <a href='https://github.com/No-tengo-nombre/chirper'>GitHub repo</a>."
        self.status_lbl = QtWidgets.QLabel(status_bar_msg)
        self.statusBar().addPermanentWidget(self.status_lbl)

    def closeEvent(self, a0: QtGui.QCloseEvent) -> None:
        """Handles confirmation when exiting."""
        reply = QtWidgets.QMessageBox.question(
            self,
            "Quit",
            "Are you sure you want to exit the program?",
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
            QtWidgets.QMessageBox.No
        )
        if reply == QtWidgets.QMessageBox.Yes:
            a0.accept()
        else:
            a0.ignore()

    def log_msg(self, msg) -> None:
        logging.info(msg)


########################################################################################################################
# |||||||||||||||||||||||||||||||||||||||| Main widget for the window |||||||||||||||||||||||||||||||||||||||||||||||| #
########################################################################################################################


class ChirperMainWidget(QtWidgets.QWidget):
    """Main content widget."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.layout = QtWidgets.QHBoxLayout()
        self.setLayout(self.layout)
        self.start()

    def start(self) -> None:
        """Initializes the main widget."""
        self.config_widget = ChirperConfigWidget(self)
        self.data_widget = ChirperDataWidget(self)

        self.layout.addWidget(self.config_widget)
        self.layout.addWidget(self.data_widget.fig, 2)

    def log_msg(self, msg):
        self.parent().log_msg(msg)

########################################################################################################################
# |||||||||||||||||||||||||||||||||||||||||||| Configuration column |||||||||||||||||||||||||||||||||||||||||||||||||| #
########################################################################################################################


class ChirperConfigWidget(QtWidgets.QWidget):
    """Widget that contains all the configurations for the app."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.layout = QtWidgets.QVBoxLayout()
        self.layout.setAlignment(QtCore.Qt.AlignTop)
        self.setLayout(self.layout)
        self.start()

    def start(self) -> None:
        """Initializes the configuration widget."""
        # Make all the options and buttons
        self.source_options = [
            "microphone",
        ]

        self.source_btns = [
            # ("On", lambda: logging.info("Source turned ON")),
            # ("Off", lambda: logging.info("Source turned OFF")),
            ("On", self.source_on_event),
            ("Off", self.source_off_event),
        ]

        self.types_options = [
            "spectrogram",
        ]

        self.console_options = [
            "DEBUG",
            "INFO",
            "WARNING",
            "ERROR",
            "CRITICAL",
        ]

        # Adding the widgets to the main one
        self.title_entry = self.make_options_entry(
            "<h3>Configuration</h3><hr>"
        )
        self.source_entry = self.make_options_entry(
            "Source",
            self.source_options,
            self.source_btns,
            lambda: self.source_options_event(self.source_entry),
            0,
            self.source_options_default,
        )
        self.types_entry = self.make_options_entry(
            "Visualization",
            self.types_options,
            None,
            lambda: self.type_options_event(self.types_entry),
            0,
            self.type_options_default,
        )
        self.blocksize_entry = self.make_number_config(
            "Blocksize",
            500,
            lambda: self.blocksize_event(self.blocksize_entry),
            self.blocksize_default,
        )
        self.samprate_entry = self.make_number_config(
            "Sampling rate",
            44100,
            lambda: self.samplerate_event(self.samprate_entry),
            self.samplerate_default,
        )
        self.maxtime_entry = self.make_number_config(
            "Maximum time",
            2,
            lambda: self.maxtime_event(self.maxtime_entry),
            self.maxtime_default,
        )
        self.channels_entry = self.make_number_config(
            "Channels",
            1,
            lambda: self.channels_event(self.channels_entry),
            self.channels_default,
        )
        self.console_config_entry = self.make_options_entry(
            "Console level",
            self.console_options,
            options_event=lambda: self.console_options_event(
                self.console_config_entry),
            default_index=2,
        )

        # Making the console boxes
        self.output_console_box = self.make_console_box()
        self.input_console_box = self.make_input_console_box()

        # We format the output console as a logger
        self.output_console_box.setFormatter(
            logging.Formatter("%(levelname)s - %(message)s"))
        self.output_console_box.setReadOnly(True)
        logging.getLogger().addHandler(self.output_console_box)

        self.input_console_box.setPlaceholderText("Chirper console")

        # Adding a 'clear console' button
        console_clear_btn = QtWidgets.QPushButton("Clear console")
        console_clear_btn.clicked.connect(
            lambda: self.output_console_box.setText(""))
        self.console_config_entry.layout().addWidget(console_clear_btn)

    # The following are all handlers for all the events
    def source_options_event(self, entry) -> None:
        option_index = entry.options_box.currentIndex()
        self.source = self.source_options[option_index]

    def source_options_default(self, index) -> None:
        self.source = self.source_options[index]

    def type_options_event(self, entry) -> None:
        option_index = entry.options_box.currentIndex()
        self.request_type = self.types_options[option_index]

    def type_options_default(self, index) -> None:
        self.request_type = self.types_options[index]

    def blocksize_event(self, entry) -> None:
        value = entry.input_box.text()
        try:
            self.blocksize = int(value)
            logging.info(f"Set blocksize to {self.blocksize}")
        except ValueError:
            logging.warning(f"Can't understand {value} as int.")

    def samplerate_event(self, entry) -> None:
        value = entry.input_box.text()
        try:
            self.samplerate = int(value)
            logging.info(f"Set samplerate to {self.samplerate}")
        except ValueError:
            logging.warning(f"Can't understand {value} as int.")

    def maxtime_event(self, entry) -> None:
        value = entry.input_box.text()
        try:
            self.max_time = float(value)
            logging.info(f"Set max time to {self.max_time}")
        except ValueError:
            logging.warning(f"Can't understand {value} as float.")

    def channels_event(self, entry) -> None:
        value = entry.input_box.text()
        try:
            self.channels = int(value)
            logging.info(f"Set channels to {self.channels}")
        except ValueError:
            logging.warning(f"Can't understand {value} as int.")

    def blocksize_default(self, value) -> None:
        try:
            self.blocksize = int(value)
            logging.info(f"Set blocksize to {self.blocksize}")
        except ValueError:
            logging.warning(f"Can't understand {value} as int.")

    def samplerate_default(self, value) -> None:
        try:
            self.samplerate = int(value)
            logging.info(f"Set samplerate to {self.samplerate}")
        except ValueError:
            logging.warning(f"Can't understand {value} as int.")

    def maxtime_default(self, value) -> None:
        try:
            self.max_time = float(value)
            logging.info(f"Set max time to {self.max_time}")
        except ValueError:
            logging.warning(f"Can't understand {value} as float.")

    def channels_default(self, value) -> None:
        try:
            self.channels = int(value)
            logging.info(f"Set channels to {self.channels}")
        except ValueError:
            logging.warning(f"Can't understand {value} as int.")

    def source_on_event(self) -> None:
        # Acquire the data widget from the parent and tell it to start
        data_widget = self.parent().data_widget
        data_widget.set_source(self.source)
        data_widget.send_start_request()

        # Create the adequate fetch request and start fetching
        data_widget.fetch_request["request_type"] = self.request_type
        data_widget.fetch_request["blocksize"] = self.blocksize
        data_widget.fetch_request["samplerate"] = self.samplerate
        data_widget.fetch_request["max-time"] = self.max_time
        data_widget.fetch_request["channels"] = self.channels

        # This method loops until the user sends a stop request
        data_widget.start_fetch()

    def source_off_event(self) -> None:
        data_widget = self.parent().data_widget

        # The procedure for stopping the source is
        #  - Stop fetching
        #  - Send a stop request
        #  - Clearing the image (this last one can be omitted)
        # It is important to do the first two in that order
        data_widget.stop_fetch()
        data_widget.send_stop_request()
        data_widget.clear_image()

    def console_options_event(self, entry) -> None:
        options = [
            logging.DEBUG,
            logging.INFO,
            logging.WARNING,
            logging.ERROR,
            logging.CRITICAL,
        ]
        option_index = entry.options_box.currentIndex()
        logging.getLogger().setLevel(options[option_index])

    def make_options_entry(self, msg=None, options=None, btns=None,
                           options_event=None, default_index=None,
                           default_setter=None) -> QtWidgets.QWidget:
        # The container widget for an entry is an HBox
        entry = QtWidgets.QWidget()
        entry_layout = QtWidgets.QHBoxLayout()
        entry_layout.setAlignment(QtCore.Qt.AlignLeft)
        entry.setLayout(entry_layout)

        if msg is not None:
            # If a message is specified we make a label with it
            entry.text_box = QtWidgets.QLabel(msg)
            entry_layout.addWidget(entry.text_box)

        if options is not None:
            # The options are given in a combo box
            entry.options_box = QtWidgets.QComboBox()
            entry.options_box.addItems(options)
            if default_index is not None:
                entry.options_box.setCurrentIndex(default_index)
                if default_setter is not None:
                    default_setter(default_index)
            if options_event is not None:
                # We add the event handler for changing the option (if
                # it is specified)
                entry.options_box.activated.connect(options_event)
            entry_layout.addWidget(entry.options_box)

        if btns is not None:
            entry.btns = []
            for btn, *action in btns:
                option = QtWidgets.QRadioButton(btn)
                if action is not None:
                    option.clicked.connect(*action)
                entry_layout.addWidget(option)
                entry.btns.append(option)

        self.layout.addWidget(entry)
        return entry

    def make_number_config(self, msg, default=None, enter_event=None,
                           default_setter=None) -> QtWidgets.QWidget:
        # The container widget for an entry is an HBox
        entry = QtWidgets.QWidget()
        entry_layout = QtWidgets.QHBoxLayout()
        entry_layout.setAlignment(QtCore.Qt.AlignLeft)
        entry.setLayout(entry_layout)

        if msg is not None:
            # If a message is specified we make a label with it
            entry.text_box = QtWidgets.QLabel(msg)
            entry_layout.addWidget(entry.text_box)

        # For a number configuration, the entry box is a QLineEdit
        entry.input_box = QtWidgets.QLineEdit()
        if default is not None:
            # We set the default number if it is given
            entry.input_box.setText(str(default))
            if default_setter is not None:
                default_setter(default)
        if enter_event is not None:
            # We add an event that happens when we press the Enter key
            # within the text box
            entry.input_box.returnPressed.connect(enter_event)
        entry_layout.addWidget(entry.input_box)

        self.layout.addWidget(entry)
        return entry

    def make_console_box(self) -> QtWidgets.QTextEdit:
        console = ConsoleBox()
        # Courier New is the "computer looking" font
        console.setFont(QtGui.QFont("Courier New"))
        self.layout.addWidget(console)
        return console

    def make_input_console_box(self) -> QtWidgets.QLineEdit:
        console = InputConsoleBox()
        # Courier New is the "computer looking" font
        console.setFont(QtGui.QFont("Courier New"))
        self.layout.addWidget(console)
        return console

    def log(self, msg) -> None:
        self.output_console_box.append(msg)

########################################################################################################################
# ||||||||||||||||||||||||||||||||||||||||| Data visualization column |||||||||||||||||||||||||||||||||||||||||||||||| #
########################################################################################################################


class ChirperDataWidget(QtWidgets.QWidget):
    """Widget that contains the data visualization."""

    def __init__(self, *args, **kwargs) -> None:
        super(QtWidgets.QWidget, self).__init__(*args, **kwargs)
        self.gui = GuiInterface()
        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)
        self.start(*args, **kwargs)

    def start(self, *args, cmap=None, **kwargs) -> None:
        """Initialize the widget."""

        # We make the default requests
        self.start_request = {
            "request_type": "start",
        }
        self.fetch_request = {}
        self.stop_request = {
            "request_type": "stop",
        }

        # This attribute is the one that will be shown on the screen
        self.fig = pg.ImageView()

        # Set the colormap if it is specified, otherwise set it to plasma
        if cmap is not None:
            self.set_cmap(cmap)
        else:
            self.set_cmap("plasma")

        self.layout.addWidget(self.fig)

    def clear_image(self) -> None:
        """Clear the content of the main figure."""
        self.send_request({
            "request_type": "clear",
            "source": self.source,
        })
        self.fig.clear()

    @dispatch(str)
    def set_cmap(self, cmap: str) -> None:
        self.fig.setColorMap(pg.colormap.get(cmap))

    def set_source(self, source) -> None:
        """Sets the source of the widget."""
        self.source = source
        self.start_request["source"] = source
        self.stop_request["source"] = source
        self.fetch_request["source"] = source

    def start_fetch(self, interval=20) -> None:
        """Starts the loop to fetch the data."""
        self.timer = pg.Qt.QtCore.QTimer()
        self.timer.setInterval(interval)
        self.timer.timeout.connect(self.update_plot_data)
        self.timer.start()

    def stop_fetch(self) -> None:
        """Stops the data fetch loop."""
        self.timer.stop()

    def update_plot_data(self) -> None:
        self.values = self.send_fetch_request()
        self.fig.setImage(self.values.values)

    def send_request(self, request: dict) -> Signal:
        """Sends the given request to the API."""
        logging.info(f"Send request {request}")
        return self.gui.make_request(request)

    def send_start_request(self) -> Signal:
        """Sends the contained start request to the API."""
        logging.info(f"Send start request {self.start_request}")
        return self.gui.make_request(self.start_request)

    def send_fetch_request(self) -> Signal:
        """Sends the contained fetch request to the API."""
        logging.info(f"Send fetch request {self.fetch_request}")
        return self.gui.make_request(self.fetch_request)

    def send_stop_request(self) -> Signal:
        """Sends the contained stop request to the API."""
        logging.info(f"Send stop request {self.stop_request}")
        return self.gui.make_request(self.stop_request)

########################################################################################################################
# |||||||||||||||||||||||||||||||||||||||||||||||| Console Box ||||||||||||||||||||||||||||||||||||||||||||||||||||||| #
########################################################################################################################


class ConsoleBox(QtWidgets.QTextEdit, logging.Handler):
    """Widget for a read only console."""

    def emit(self, record):
        msg = self.format(record)
        self.append(msg)

    def append(self, text: str) -> None:
        """Appends the given text with a >>> in the beginning."""
        return super().append(f">>> {text}")


class InputConsoleBox(QtWidgets.QLineEdit):
    """Widget for a console that allows the user to send data."""
    pass

########################################################################################################################
# ||||||||||||||||||||||||||||||||||||||||||||||| Main function |||||||||||||||||||||||||||||||||||||||||||||||||||||| #
########################################################################################################################


def main():
    app = QtWidgets.QApplication(sys.argv)
    ex = ChirperApp()
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



import numpy as np

from ..sgn import Signal1
from ..config import PM_MODULATION, HERTZ


def pm_modulation(signal1: Signal1, *args, method=PM_MODULATION, hertz=HERTZ, **kwargs) -> Signal1:
    """Applies PM modulation to the given one dimensional signal.

    The currently available methods for modulation are:
     - Traditional

    Parameters
    ----------
    signal1 : Signal1
        One dimensional signal to modulate.
    carrier_freq : float
        Frequency of the carrier wave.
    carrier_amp : float
        Amplitude of the carrier wave.
    method : {"trad"}, optional
        Method used for the modulation, by default PM_MODULATION.
    hertz : bool, optional
        Whether the frequency is given in Hertz, by default HERTZ.

    Returns
    -------
    Signal1
        Modulated one dimensional signal.
    """
    return PM_MODULATION_METHODS[method](signal1, *args, hertz, **kwargs)


def _trad_modulation(signal1: Signal1, carrier_freq, carrier_amp, hertz):
    copy = signal1.clone()
    axis = copy.axis
    freq = 2 * np.pi * carrier_freq if hertz else carrier_freq
    values = carrier_amp * np.sin(freq * axis + copy.values)
    return Signal1(axis, values)


PM_MODULATION_METHODS = {
    "trad": _trad_modulation,
}



import unittest

from src.chirper.sgn import Signal1


class TestSignal(unittest.TestCase):
    def setUp(self):
        self.signal1 = Signal1(
            [i for i in range(100)],
            [i ** 2 for i in range(100)]
        )
        self.signal2 = Signal1.from_function(
            [i for i in range(100)],
            lambda x: x ** 2
        )
        self.signal3 = Signal1(
            [i for i in range(100)],
            [100 for _ in range(100)]
        )
        self.signal4 = Signal1(
            [2 * i for i in range(50)],
            [100 for _ in range(50)]
        )
        self.signal5 = Signal1.from_function(
            [2 * i for i in range(50)],
            lambda x: 100 + x - x
        )
        self.signal6 = Signal1.from_function(
            [i for i in range(100)],
            lambda x: 200 + x - x
        )
        self.signal7 = Signal1.from_function(
            [i for i in range(100)],
            lambda x: (x - 50) ** 3
        )
        self.signal8 = Signal1.from_function(
            [1.5 * i for i in range(100)],
            lambda x: x ** 3
        )

    def test_creation(self):
        self.assertEqual(self.signal1, self.signal2,
                         "Time signal creation test failed")

    def test_indexing(self):
        self.assertEqual(
            45 ** 3, self.signal8(45), "Continous time signal indexing test failed")
        for t in self.signal8.axis:
            self.assertEqual(
                t ** 3, self.signal8(t), "Continous time signal indexing test failed")

    def test_addition(self):
        exp_signal = Signal1.from_function(
            [i for i in range(100)],
            lambda x: 2 * (x ** 2)
        )
        real_signal = self.signal1 + self.signal2
        self.assertEqual(exp_signal, real_signal,
                         "Time signal addition test failed")

    def test_subtraction(self):
        exp_signal1 = Signal1.from_function(
            [i for i in range(100)],
            lambda x: x - x
        )
        exp_signal2 = Signal1.from_function(
            [i for i in range(100)],
            lambda x: (x ** 2) - 100
        )
        self.assertEqual(exp_signal1, self.signal1 - self.signal2,
                         "Time signal subtraction test failed")
        self.assertEqual(exp_signal2, self.signal1 - self.signal3,
                         "Time signal subtraction test failed")

    def test_multiplication(self):
        exp_signal = Signal1.from_function(
            [i for i in range(100)],
            lambda x: x ** 4
        )
        real_signal = self.signal1 * self.signal2
        self.assertEqual(exp_signal, real_signal,
                         "Time signal multiplication test failed")

    def test_division(self):
        exp_signal1 = Signal1.from_function(
            [i for i in range(1, 100)],
            lambda x: x / x
        )
        exp_signal2 = Signal1.from_function(
            [i for i in range(100)],
            lambda x: (x ** 2) / 100
        )
        sign1_cut = Signal1(self.signal1.axis[1:], self.signal1.values[1:])
        sign2_cut = Signal1(self.signal2.axis[1:], self.signal2.values[1:])
        div = (self.signal1 / self.signal2)

        real_signal1 = sign1_cut / sign2_cut
        real_signal2 = Signal1(div.axis[1:], div.values[1:])
        self.assertEqual(exp_signal1, real_signal1,
                         "Time signal division test failed")
        self.assertEqual(exp_signal1, real_signal2,
                         "Time signal division test failed")
        self.assertEqual(exp_signal2, self.signal1 /
                         self.signal3, "Time signal division test failed")

    def test_equality(self):
        self.assertEqual(self.signal1, self.signal1,
                         "Time signal equality test failed")
        self.assertEqual(self.signal2, self.signal2,
                         "Time signal equality test failed")

    def test_abs(self):
        exp_signal = Signal1.from_function(
            [i for i in range(100)],
            lambda x: abs((x - 50) ** 3)
        )
        self.assertEqual(self.signal1, abs(self.signal1),
                         "Time signal absolute value test failed")
        self.assertEqual(exp_signal, abs(self.signal7),
                         "Time signal absolute value test failed")


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .input_source import InputSource
    from .data_process import DataProcess
    from .data_handler import DataHandler


class ChirpType:
    def get_processed(self, data_process: DataProcess, data, **kwargs):
        """Gets data processed by a data processor."""
        return None

    def get_handled(self, data_handler: DataHandler, signal, **kwargs):
        """Gets a signal handled by a data handler."""
        return None

    def fetch(self, input_source: InputSource, chirp_source: ChirpSource, **kwargs):
        return None


class ChirpSource:
    def get_fetched(self, input_source: InputSource, **kwargs):
        """Gets fetched by an input source."""
        return None


class Chirp:
    """Class for a request to the package."""

    def __init__(self, request_type: ChirpType, source: ChirpSource) -> None:
        self.request_type = request_type
        self.source = source

    def __str__(self) -> str:
        return f"request_type: {self.request_type} - source: {self.source}"




# import numpy as np
# from tqdm import tqdm

# class Z_transform(Transform):
#     def __init__(self, target):
#         super().__init__(target)
#         self.methods = {
#             "dzt": self.calculate_dzt,
#         }

#     def calculate(self, method=Z_METHOD, *args, **kwargs):
#         return self.methods[method](*args, **kwargs)

#     def calculate_dzt(self, R):
#         N = len(self.signal)
#         rho = np.empty((N, N))
#         for k in tqdm(range(N), desc="Calculating Z transform"):
#             for m in range(N):
#                 r_vec = np.array([(R * m / N) ** (-i) for i in range(N)])
#                 xi_vec = np.array([self.signal.values[i] * np.exp(-1j * 2 * np.pi * (k / N) * i) for i in range(N)])
#                 rho[k, m] = np.dot(r_vec, xi_vec)
#         self.values = rho
#         return self.values



from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np

from ..sgn import Signal1
if TYPE_CHECKING:
    from . import GuiInterface
    from .chirp import Chirp


class DataProcess:
    def __init__(self, api: GuiInterface) -> None:
        self.api = api

    def process(self, data, request: Chirp, **kwargs):
        return request.request_type.get_processed(self, data, **kwargs)

    def process_spectrogram(self, data, **kwargs):
        samp_freq = self.api.samplerate
        values = data.mean(axis=1)
        return Signal1.from_freq(values, samp_freq)



from __future__ import annotations
from typing import TYPE_CHECKING
import logging
import numpy as np

from ..sgn import Signal1, Signal2
from ..transforms import fourier
if TYPE_CHECKING:
    from . import GuiInterface
    from .chirp import Chirp


class DataHandler:
    def __init__(self, api: GuiInterface) -> None:
        self.api = api
        self.values = None

    def handle(self, signal: Signal1, request: Chirp, **kwargs):
        return request.request_type.get_handled(self, signal, **kwargs)

    def handle_spectrogram(self, signal: Signal1, half=True, positive_half=True, max_time=5, **kwargs):
        if half:
            fourier_signal = fourier.f1(signal).half(positive_half)
        else:
            fourier_signal = fourier.f1(signal)

        # This runs on the first fetch request
        if self.values is None:
            return self._handle_spectrogram_empty(fourier_signal, max_time, **kwargs)

        # If the data was fetched before, `self.values` will not
        # be `None`
        else:
            return self._handle_spectrogram_notempty(fourier_signal, max_time, **kwargs)

    def _handle_spectrogram_empty(self, fourier_signal, max_time, **kwargs):
        self.dt = self.api.blocksize / self.api.samplerate

        # Accessing [:, None] turns the array into a column vector.
        # Then we take the transpose so that it has the proper dimensions
        self.values = Signal2(
            [0],
            fourier_signal.axis,
            fourier_signal.values[:, None].T,
        )
        assert self.values.is_valid(), "Something went wrong"
        self.permanent_values = self.values.clone()
        return self.values.abs()

    def _handle_spectrogram_notempty(self, fourier_signal, max_time, **kwargs):
        self.values.ax0 = np.append(
            self.values.ax0,
            self.values.ax0[-1] + self.dt,
        )
        self.values.values = np.vstack(
            (self.values.values, fourier_signal.values)
        )

        # The permanent values are the same as the values, except
        # they are not cut when they exceed the length
        self.permanent_values.ax0 = np.append(
            self.permanent_values.ax0,
            self.permanent_values.ax0[-1] + self.dt,
        )
        self.permanent_values.values = np.vstack(
            (self.permanent_values.values, fourier_signal.values)
        )

        if self.values.ax0_span() > max_time:
            end_time = self.values.ax0[-1]
            self.values = self.values.get_ax0(end_time - max_time)

        assert self.values.is_valid(), "Something went wrong"
        return self.values.abs()

    def clear_data(self, signal: Signal1, **kwargs):
        logging.info("Clearing data in data handler")
        self.values = None
